package net.chlup.mybrother;

import java.util.*;

import android.app.Service;
import android.app.AlarmManager;
import android.app.NotificationManager;
import android.app.PendingIntent;
import android.app.Notification;

import android.os.Bundle;
import android.os.Environment;
import android.os.IBinder;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.BroadcastReceiver;

import android.location.*;
import android.telephony.TelephonyManager;
import android.util.Base64;
import android.util.Log;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyManagementException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.text.SimpleDateFormat;

import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.NameValuePair;
import org.apache.http.client.ClientProtocolException;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.client.DefaultHttpRequestRetryHandler;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.util.EntityUtils;

public class GPS extends Service {
   /* Location manager params */
   LocationManager lm;
   LocListen ll;
   AlarmManager am;
   PendingIntent amintent;
   LocAlarm recvr;
   PendingIntent tointent;
   LocTimeout recvTimeout;
   boolean twiceTimeout;
   long timeout;

   /* Prepare connection and request */
   DefaultHttpClient httpclient = null; // presunuto sem nahoru z "postLocation()", aby fungoval keep-alive pri rychlejsim zasilani souradnic
   int waiting_for_first_send_data = 1;

   /* Pole se zjistenymi udaji */
   long numberPoint = 0;
   long numberPointSends = 0;
   boolean firstSend = true; // zajisti ze prvi zjisteni polohy bude odeslano okamzite (bez ohledu na to v jakych davkach se ma poloha zasilat)
   List<NameValuePair> nameValuePairs;

   /* Date formatting */
   SimpleDateFormat dateformatter;
   SimpleDateFormat dateformattercsv;

   /* Notification */
   NotificationManager notman;
   Notification notif;
   PendingIntent notintent;

   /* RPC */
   LocBinder binder;
   public LocIF rpc_if;
   URL target_url;

   /* Our position data from last read */
   Location location;
   long locationTimePrevious = 0;

   /* Battery info */
   BatteryState bat_rcvr;
   int bat_level;
   boolean charger;

   /* Prefs */
   Preferences prefs;

   /* Storage PATH */
   String storagepath = Environment.getExternalStorageDirectory().toString();

   /* CSV */
   String filesavepathname = storagepath + "/mybrother";
   String filesavename = null;
   String[] csvcolsnames = {
         "TYPE","TIME","LAT","LON","ALT","BEARING","ACCURACY","SPEED","NAME","DESCRIPTION","SEGMENT", // cast ktera je v CSV u MyTracks
         "PROV","BAT","CHAR" // cast, kterou jsem si doplnil a neni v CSV u MyTracks
   };

   @Override public void onCreate() {
      super.onCreate();
      Log.i(this.getClass().getName(), "GPS Service onCreate.");

      /* Create formatter */
      dateformatter = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SS'Z'");
      dateformatter.setTimeZone(new SimpleTimeZone(0, "UTC"));
      dateformattercsv = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'");
      dateformattercsv.setTimeZone(new SimpleTimeZone(0, "UTC"));

      setupLocationListener();

      setupAlarms();

      setupBattery();

      /* Get prefs */
      this.prefs = new Preferences(this);
      this.prefs.load();

      setupHttpConnection();

      reconfigure();

      /* Create binder */
      this.binder = new LocBinder(this);
   }

   @Override public void onDestroy() {
      super.onDestroy();
      Log.i(this.getClass().getName(), "GPS Service onDestroy.");

      /* Remove alarms and unhook locator */
      this.am.cancel(this.amintent);
      this.am.cancel(this.tointent);
      this.lm.removeUpdates(this.ll);
      unregisterReceiver(this.recvr);
      unregisterReceiver(this.bat_rcvr);
      unregisterReceiver(this.recvTimeout);

      if (this.notman != null) {
         /* Remove notification */
         this.notman.cancelAll();
      }
   }


   @Override public IBinder onBind(Intent i) {
      return this.binder;
   }

   @Override public boolean onUnbind(Intent i) {
      this.rpc_if = null;
      return false;
   }


   private void startLocator() throws IllegalArgumentException {
      if (this.prefs.provider == 1)
         this.lm.requestLocationUpdates(LocationManager.GPS_PROVIDER, 0, 0, this.ll);
      else
         this.lm.requestLocationUpdates(LocationManager.NETWORK_PROVIDER, 0, 0, this.ll);
   }

   public void triggerUpdate() {
      if (this.prefs.continous_mode) {
         /* If we are in continous mode just send last location */
         locationUpdate();
      }
      else {
         /* Polling mode */

         this.lm.removeUpdates(this.ll);

         try {
            startLocator();
         }
         catch (IllegalArgumentException e) {
            Log.e(this.getClass().getName(), "MyBrother: "+e.toString());
            return;
         }

         /* Start timeout alarm */
         this.timeout = System.currentTimeMillis();
         this.timeout += this.prefs.gps_timeout;
         this.timeout += 100; /* delay a bit to avoid a race */
         this.am.setRepeating(
               AlarmManager.RTC_WAKEUP,
               this.timeout,
               this.prefs.gps_timeout,
               this.tointent
         );
         if (this.prefs.provider == 1)
            this.twiceTimeout = true;
         else
            this.twiceTimeout = false;
      }
   }

   public void triggerUpdateButtonUpdateNow() {
      firstSend = true;
      if (this.prefs.continous_mode) {
         /* Reset update alarms */
         this.am.setRepeating(
               AlarmManager.RTC_WAKEUP,
               System.currentTimeMillis() + 1000,
               this.prefs.update_interval,
               this.amintent
         );
      }
      else {
         // pro cudl Update Now resim zatim situaci jen pro continous_mode (viz vyse)
         // mozna bych zde mohl zavolat primo triggerUpdate()
      }
   }

   private void setupHttpConnection() {
      this.httpclient = new DefaultHttpClient();
      this.httpclient.setHttpRequestRetryHandler(new DefaultHttpRequestRetryHandler(0, false)); // retry jen 0x (napriklad je nedostupna sit) (pokud bych to neuvedl, tak defaultne je tusim retry 3x)
      this.nameValuePairs = new ArrayList<NameValuePair>(2);
      if(this.prefs.own_key_store) {
         try {
            KeyStore trustStore = KeyStore.getInstance(KeyStore.getDefaultType());
            FileInputStream instream = new FileInputStream(new File(storagepath + "/security/cacerts.bks"));
            trustStore.load(instream, "changeit".toCharArray());
            instream.close();
            SSLSocketFactory socketFactory = new SSLSocketFactory(trustStore);
            Scheme sch = new Scheme("https", socketFactory, 443);
            this.httpclient.getConnectionManager().getSchemeRegistry().register(sch);
         }
         catch (KeyStoreException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (FileNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (NoSuchAlgorithmException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (CertificateException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (KeyManagementException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
         catch (UnrecoverableKeyException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
            httpclient = null;
         }
      }
   }

   private void setupLocationListener() {
      /* Set up the position listener */
      this.ll = new LocListen();
      this.lm = (LocationManager)getSystemService(Context.LOCATION_SERVICE);
   }

   private void setupNotif() {
      if (this.prefs.show_in_notif_bar) {
         /* Show state in notif bar */

         if (this.notman != null) {
            /* Already set up */
            return;
         }

         /* Set up a persistent notification */
         this.notman = (NotificationManager)
         getSystemService(Context.NOTIFICATION_SERVICE);
         this.notif = new Notification(
               R.drawable.notif_icon,
               null, // neuvadim zadny text, aby na status baru bylo okamzite videt, ze ikona gps naskocila
               System.currentTimeMillis()
         );
         this.notif.flags = Notification.FLAG_ONGOING_EVENT;
         this.notintent =
         PendingIntent.getActivity(
               this,
               0,
               new Intent(this, MyBrother.class),
               0
         );
         this.notif.setLatestEventInfo(
               this, getString(R.string.app_name),
               getString(R.string.notifiWaitingForInitialLocation),
               notintent
         );
         this.notman.notify(0, notif);
      }
      else {
         if (this.notman != null) {
            /* Remove notification */
            this.notman.cancelAll();
            this.notman = null;
         }
      }
   }

   private void setupAlarms() {
      /* Prepare alarm manager */
      this.am = (AlarmManager)getSystemService(Context.ALARM_SERVICE);

      /* Setup location update alarm */
      this.recvr = new LocAlarm();
      registerReceiver(
            this.recvr,
            new IntentFilter(LocAlarm.class.toString()),
            null,
            null
      );
      Intent i = new Intent(LocAlarm.class.toString());
      this.amintent = PendingIntent.getBroadcast(this, 0, i, 0);

      /* Setup timeout alarm */
      this.recvTimeout = new LocTimeout();
      registerReceiver(
            this.recvTimeout,
            new IntentFilter(LocTimeout.class.toString()),
            null,
            null
      );
      i = new Intent(LocTimeout.class.toString());
      this.tointent = PendingIntent.getBroadcast(this, 0, i, 0);
   }

   private void setupBattery() {
      /* Setup location update alarm */
      this.bat_rcvr = new BatteryState();
      registerReceiver(
            this.bat_rcvr,
            new IntentFilter(Intent.ACTION_BATTERY_CHANGED),
            null,
            null
      );
   }


   public void doTimeout() {
      if (System.currentTimeMillis() > this.timeout) {
         Log.i(this.getClass().getName(), "Doing timeout");
         if (this.twiceTimeout) {
            Log.i(this.getClass().getName(), "Switching locator");
            this.twiceTimeout = false;
            this.timeout =
               System.currentTimeMillis() + this.prefs.gps_timeout;
            try {
               this.lm.requestLocationUpdates(
                     LocationManager.NETWORK_PROVIDER,
                     0,
                     0,
                     this.ll
               );
            }
            catch (IllegalArgumentException e) {
               Log.e(this.getClass().getName(), "Timeout: "+e.toString());
               return;
            }
         }
         else {
            /* Timeout reached */
            this.lm.removeUpdates(this.ll);
            this.am.cancel(this.tointent);
         }
      }
   }


   public void reloadPrefs() {
      this.prefs.load();
      reconfigure();
   }

   private void reconfigure() {
      Log.i(this.getClass().getName(), "doing reconfig");
      /* Update the request times */
      this.lm.removeUpdates(ll);

      /* Reset update alarms */
      this.am.setRepeating(
            AlarmManager.RTC_WAKEUP,
            System.currentTimeMillis() + 1000,
            this.prefs.update_interval,
            this.amintent
      );

      /* Fix notifs */
      setupNotif();

      /* New file */
      if(this.prefs.file_save) {
         File filepath = new File(filesavepathname);
         if(!filepath.exists()) {
            filepath.mkdir();
         }
         SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd_HHmmss");
         sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
         String datetimestr = sdf.format(new Date());
         filesavename = filesavepathname + "/" + datetimestr + ".csv";
         try {
            BufferedWriter fileout = new BufferedWriter(new FileWriter(filesavename));
            for (String colname : csvcolsnames) {
               if(colname != "TYPE") {
                  fileout.append(',');
               }
               fileout.append(stringCsv(colname));
            }
            fileout.append('\n');
            for (String colname : csvcolsnames) {
               if(colname != "TYPE") {
                  fileout.append(',');
               }
               if(colname == "TYPE") {
                  fileout.append(stringCsv("TRACK"));
               }
               else if(colname == "NAME") {
                  if(this.prefs.http_user != null) {
                     fileout.append(stringCsv(this.prefs.http_user));
                  }
                  else {
                     fileout.append(stringCsv(datetimestr));
                  }
               }
            }
            fileout.append('\n');
            fileout.flush();
            fileout.close();
         }
         catch (IOException e) {
            Log.e(this.getClass().getName(), "Can't first write to CSV file: " + e.toString());
         }
      }

      /* Set URL */
      this.target_url = null;
      try {
         this.target_url = new URL(this.prefs.target_url);
      }
      catch (MalformedURLException e) {
         Log.e(this.getClass().getName(), "target_url: "+e.toString());
         this.target_url = null;
      }

      /* For continous mode, start updating */
      if (this.prefs.continous_mode) {
         try {
            startLocator();
         }
         catch (IllegalArgumentException e) {
            Log.e(
                  this.getClass().getName(), 
                  "Can't start locator in continous mode: " +
                  e.toString()
            );
            return;
         }
      }
   }

   /* Send a request to the URL and post some data */
   protected void postLocation() {

      /* No url, don't do anything */
      if (this.target_url == null)
         return;

      if(
            this.prefs.show_in_notif_bar &&
            this.prefs.http_resp_in_notif_bar
      ) {
         this.setupNotif();
         if(
               this.notif != null &&
               waiting_for_first_send_data==1
         ) {
            this.notif.when = System.currentTimeMillis();
            this.notif.setLatestEventInfo(
                  this,
                  getString(R.string.app_name),
                  getString(R.string.notifiWaitingForFirstSendDataToServer),
                  this.notintent
            );
            this.notman.notify(0, this.notif);
            waiting_for_first_send_data = 0;
         }
      }

      /* Build request data */

      ++numberPoint;
      ++numberPointSends;
      String numberPointText;
      if(numberPoint==1) {
         numberPointText="";
      }
      else {
         numberPointText=String.valueOf(numberPoint);
      }

      Map<String, String> csv = new HashMap<String, String>(); 

      csv.put("TYPE", "P");
      nameValuePairs.add(new BasicNameValuePair("latitude"+numberPointText, String.format(Locale.US, "%.6f",this.location.getLatitude())));
      csv.put("LAT", String.format(Locale.US, "%.6f",this.location.getLatitude()));
      nameValuePairs.add(new BasicNameValuePair("longitude"+numberPointText, String.format(Locale.US, "%.6f",this.location.getLongitude())));
      csv.put("LON", String.format(Locale.US, "%.6f",this.location.getLongitude()));
      nameValuePairs.add(new BasicNameValuePair("altitude"+numberPointText, String.valueOf(this.location.getAltitude())));
      csv.put("ALT", String.valueOf(this.location.getAltitude()));
      nameValuePairs.add(new BasicNameValuePair("accuracy"+numberPointText, String.valueOf(this.location.getAccuracy())));
      csv.put("ACCURACY", String.valueOf(this.location.getAccuracy()));

      nameValuePairs.add(new BasicNameValuePair("provider"+numberPointText, this.location.getProvider()));
      csv.put("PROV", String.valueOf(this.location.getProvider()));

      nameValuePairs.add(new BasicNameValuePair("bearing"+numberPointText, String.valueOf(this.location.getBearing())));
      csv.put("BEARING", String.valueOf(this.location.getBearing()));
      nameValuePairs.add(new BasicNameValuePair("speed"+numberPointText, String.valueOf(this.location.getSpeed())));
      csv.put("SPEED", String.valueOf(this.location.getSpeed()));

      Date date = new Date(this.location.getTime());
      nameValuePairs.add(new BasicNameValuePair("time"+numberPointText, this.dateformatter.format(date)));
      csv.put("TIME", this.dateformattercsv.format(date));

      /* Add battery status */
      String chargerval;
      if (this.charger) {
         chargerval = "1";
      }
      else {
         chargerval = "0";
      }
      if (this.prefs.send_batt_status) {
         nameValuePairs.add(new BasicNameValuePair("battlevel"+numberPointText, String.valueOf(this.bat_level)));
         nameValuePairs.add(new BasicNameValuePair("charging"+numberPointText, chargerval));
      }
      csv.put("BAT", String.valueOf(this.bat_level));
      csv.put("CHAR", chargerval);

      /* Add secret if configured */
      if (this.prefs.secret != null) {
         nameValuePairs.add(new BasicNameValuePair("secret"+numberPointText, this.prefs.secret));
      }

      /* Add device id */
      if (this.prefs.send_devid) {
         TelephonyManager tm = (TelephonyManager)getSystemService(TELEPHONY_SERVICE);
         nameValuePairs.add(new BasicNameValuePair("deviceid"+numberPointText, tm.getDeviceId()));
      }

      /* Add subscriber id */
      if (this.prefs.send_subscrid) {
         TelephonyManager tm = (TelephonyManager)getSystemService(TELEPHONY_SERVICE);
         nameValuePairs.add(new BasicNameValuePair("subscriberid"+numberPointText, tm.getSubscriberId()));
         if(tm.getPhoneType()==TelephonyManager.PHONE_TYPE_GSM) {
            nameValuePairs.add(new BasicNameValuePair("networkoperatorname"+numberPointText, tm.getNetworkOperatorName()));
            nameValuePairs.add(new BasicNameValuePair("networktype"+numberPointText, String.valueOf(tm.getNetworkType())));
         }
      }

      csv.put("SEGMENT", "0"); // nevim co je SEGMENT, ale v CSV v MyTracks to je a je to vyplneno u zaznamenanych bodu hodnotou "0"

      /* pripadny zapis do souboru CSV */
      if(this.prefs.file_save) {
         try {
            BufferedWriter fileout = new BufferedWriter(new FileWriter(filesavename, true));
            for (String colname : csvcolsnames) {
               if(colname != "TYPE") {
                  fileout.append(',');
               }
               if(csv.get(colname) != null) {
                  fileout.append(stringCsv(csv.get(colname)));
               }
            }
            fileout.append('\n');
            fileout.flush();
            fileout.close();
         }
         catch (IOException e) {
            Log.e(this.getClass().getName(), "Can't write to CSV file: " + e.toString());
         }
      }

      if(numberPointSends>=this.prefs.send_points_together || firstSend) {

         numberPointSends = 0;
         firstSend = false;

         /* reseni problemu s nepodarenou inicializaci httpclient
          *  - vetsinou pujde o chybu, ze nejsou dobre nacteny vlastni CA
          *    z lokalniho souboru */
         if(httpclient == null) {
            Log.e(this.getClass().getName(), "httpclient is null (may be problems with CA)");
            if (this.rpc_if != null)
               this.rpc_if.onError("Error: httpclient is null (may be problems with CA)");
            return;
         }

         String serverinfo="";

         HttpPost httppost = new HttpPost(this.target_url.toString());

         /* HTTP Basic Authorization */
         if (this.prefs.http_user != null && this.prefs.http_password != null) {
            // 1. varianta - je asi poctiva - prvni pozadavek je bez jmena a hesla a az si server vyzada autorizaci bude nasledovat pozadavek s uzivatelem a heslem (definice nize by vsak mela byt pripadne ve funkci setupHttpConnection())
            //httpclient.getCredentialsProvider().setCredentials(
            //        new AuthScope(AuthScope.ANY_HOST, AuthScope.ANY_PORT),
            //        new UsernamePasswordCredentials(this.prefs.http_user, this.prefs.http_password)
            //);
            // 2. varianta - poctiva - kdy jen u prvniho pozadavku je pozadavek bez jmena a hesla a az si server vyzada autorizaci bude nasledovat pozadavek s uzivatelem a heslem a u dalsich pozadavku bude jiz vzdy uzivatel a heslo
            //    - tohle jsem rozracoval u bigbrothergps-snapshot-mara03-httpclient-preemptive_authentication-for_android_2.2-TEST
            //      dle stranky zde http://hc.apache.org/httpcomponents-client-ga/tutorial/html/authentication.html,
            //      ale jsou tam jine zaludnosti a zatim to delam jinak (viz nize)
            // 3. varianta - mozna to ma nejakou zaludnost - natvrdo do hlavicky vlozim informace pro basic autorizaci a pozadavek je tak jen jeden
            String http_userpassword = this.prefs.http_user + ":" + this.prefs.http_password;
            httppost.setHeader("Authorization", "basic " + Base64.encodeToString(http_userpassword.getBytes(),Base64.NO_WRAP));
         }

         /* Connect and write */
         HttpResponse response;
         try {
            if (this.rpc_if != null)
               this.rpc_if.onError("httppost START");

            httppost.setEntity(new UrlEncodedFormEntity(nameValuePairs));
            response = this.httpclient.execute(httppost);

            HttpEntity entity = response.getEntity();
            if (entity != null) {
               serverinfo=EntityUtils.toString(entity); // tohle muze asi pripadne nahradit radek nize (pokud budes chti s obsahem co je vracen pracovat)
               if(serverinfo.length()>100) {
                  serverinfo=serverinfo.substring(0,100);
               }
               //entity.consumeContent(); // zajisti asi konzumaci obsahu s ukoncenim pozadavku - diky tomu funguje keep-alive (kdyz to tu nebylo, keep-alive nefungoval - nestacilo jen posunout definic httpclient nahoru)
            }

            if (this.rpc_if != null)
               this.rpc_if.onError("httppost END");
         }
         catch (ClientProtocolException e) {
            Log.e(this.getClass().getName(), "ClientProtocolException: " + e.toString());
            if (this.rpc_if != null)
               this.rpc_if.onError(e.toString());
            return;
         }
         catch (IOException e) {
            Log.e(this.getClass().getName(), "IOException: " + e.toString());
            if (this.rpc_if != null)
               this.rpc_if.onError(e.toString());
            return;
         }

         numberPoint = 0;
         nameValuePairs = null;
         nameValuePairs = new ArrayList<NameValuePair>(2);

         Log.i(this.getClass().getName(), "sent HTTP poke");

         /* Set notification if we have it */
         if (
               this.notif != null &&
               this.prefs.show_in_notif_bar &&
               this.prefs.http_resp_in_notif_bar
         ) {
            this.notif.when = System.currentTimeMillis();
            this.notif.setLatestEventInfo(
                  this,
                  getString(R.string.app_name),
                  "SERVER: "+serverinfo,
                  this.notintent
            );
            this.notman.notify(0, this.notif);
         }

      }

   }

   private void locationUpdate() {
      /* Location update triggered */
      Location loc = this.location;

      if (loc == null)
         return;

      // pokud je cas lokalizace stejny jako u predchozi lokalizace, tak se polohou nezabyvam - neni treba se ni zabyvat, jelikoz novy udaj o poloze neni k dispozici
      if(loc.getTime() == this.locationTimePrevious && !firstSend)
         return;
      this.locationTimePrevious = loc.getTime();

      /* Change notification */
      if (this.prefs.show_in_notif_bar && !this.prefs.http_resp_in_notif_bar) {
         this.setupNotif();
         String txt = loc.getLatitude()+", "
         +loc.getLongitude()+", "
         +(int)loc.getAccuracy()+"m";
         this.notif.when = System.currentTimeMillis();
         this.notif.setLatestEventInfo(
               GPS.this,
               getString(R.string.app_name),
               txt,
               GPS.this.notintent
         );
         this.notman.notify(0, GPS.this.notif);
      }

      /* Call to UI */
      if (this.rpc_if != null) {
         this.rpc_if.onLocation(
               loc.getProvider(),
               loc,
               this.bat_level,
               this.charger,
               this.numberPoint
         );
      }

      /* Post to server */
      this.postLocation();
   }

   /* pomocna funkce upravujici string retezec tak, aby ho bylo mozno zapsat do CSV souboru */
   private String stringCsv(String str) {
      return "\"" + str.replace("\"", "\"\"") + "\"";
   }
   
   /**************************************************************************
     * Helper classes for timed locator interaction
     *************************************************************************/
   class LocAlarm extends BroadcastReceiver {
      @Override public void onReceive(Context ctx, Intent i) {
         Log.i(this.getClass().getName(), "Alarm!");
         //android.os.SystemClock.sleep(3000); // bylo uzito pri testech (schvalne vlozene cekani simulujici, ze to dlouho trva)
         GPS.this.triggerUpdate();
      }
   }

   class LocTimeout extends BroadcastReceiver {
      @Override public void onReceive(Context ctx, Intent i) {
         if (GPS.this.prefs.continous_mode) {
            Log.i(this.getClass().getName(), "Ignored timeout");
         }
         else {
            Log.i(this.getClass().getName(), "Received Timeout!");
            GPS.this.doTimeout();
         }
      }
   }

   class BatteryState extends BroadcastReceiver {
      @Override public void onReceive(Context ctx, Intent i) {
         float level;
         level = i.getIntExtra("level", 0);
         level /= i.getIntExtra("scale", 100);
         GPS.this.bat_level = (int)(level*100);
         GPS.this.charger = i.getIntExtra("plugged",1)!=0;
         Log.i(
               this.getClass().getName(),
               String.format(
                     Locale.US,
                     "Battery state change: %d%% %b\n",
                     GPS.this.bat_level,
                     GPS.this.charger
               )
         );
      }
   }

   class LocListen implements LocationListener {
      @Override public void onProviderDisabled(String prov) {
         Log.i(this.getClass().getName(), "ProviderDisabled: "+prov);
         if (GPS.this.prefs.continous_mode)
            GPS.this.startLocator();
         else {
            /* Stop waiting for timeout since this locator is off */
            GPS.this.timeout = System.currentTimeMillis()-1;
            GPS.this.doTimeout();
         }
      }

      @Override public void onProviderEnabled(String prov) {
         Log.i(this.getClass().getName(), "ProviderEnabled: "+prov);
         if (GPS.this.prefs.continous_mode)
            GPS.this.startLocator();
         else
            GPS.this.doTimeout();
      }

      @Override public void onStatusChanged(String prov, int stat, Bundle xtra) {
         Log.i(this.getClass().getName(), "Status change: "+prov+" -> "+stat);
         if (GPS.this.rpc_if != null)
            GPS.this.rpc_if.onStateChange(prov, stat);
         if (GPS.this.prefs.continous_mode)
            GPS.this.startLocator();
         else
            GPS.this.doTimeout();
      }

      @Override public void onLocationChanged(Location loc) {
         Log.i(this.getClass().getName(), "got loc from "+loc.getProvider());
         GPS.this.location = loc;
         if (!GPS.this.prefs.continous_mode) {
            /* Stop waiting for locations. Will be restarted by alarm */
            GPS.this.lm.removeUpdates(GPS.this.ll);
            GPS.this.am.cancel(GPS.this.tointent);
            GPS.this.locationUpdate();
         }
      }
   }
}
